"""Models the state of a vehicle."""

import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List

from bimmer_connected.const import ATTR_STATE
from bimmer_connected.models import StrEnum, VehicleDataBase
from bimmer_connected.utils import to_camel_case

_LOGGER = logging.getLogger(__name__)


class LidState(StrEnum):
    """Possible states of the hatch, trunk, doors, windows, sun roof."""

    CLOSED = "CLOSED"
    OPEN = "OPEN"
    OPEN_TILT = "OPEN_TILT"
    INTERMEDIATE = "INTERMEDIATE"
    INVALID = "INVALID"
    UNKNOWN = "UNKNOWN"


class LockState(StrEnum):
    """Possible states of the door locks."""

    LOCKED = "LOCKED"
    SECURED = "SECURED"
    SELECTIVE_LOCKED = "SELECTIVE_LOCKED"
    PARTIALLY_LOCKED = "PARTIALLY_LOCKED"
    UNLOCKED = "UNLOCKED"
    UNKNOWN = "UNKNOWN"


class Lid:
    """A lid of the vehicle.

    Lids are: Doors + Trunk + Hatch
    """

    def __init__(self, name: str, state: str):
        #: name of the lid
        self.name = name
        self.state = LidState(state)

    @property
    def is_closed(self) -> bool:
        """Check if the lid is closed."""
        return self.state == LidState.CLOSED


class Window(Lid):
    """A window of the vehicle.

    A window can be a normal window of the car or the sun roof.
    """


@dataclass
class DoorsAndWindows(VehicleDataBase):
    """Provides an accessible version of `properties.doorsAndWindows`."""

    door_lock_state: LockState = LockState.UNKNOWN
    """Get state of the door locks."""

    lids: List[Lid] = field(default_factory=list)
    """All lids (doors+hood+trunk) of the car."""

    windows: List[Window] = field(default_factory=list)
    """All windows (doors+sunroof) of the car."""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Dict:
        """Parse doors and windows."""
        retval: Dict[str, Any] = {}

        if ATTR_STATE in vehicle_data:
            if "doorsState" in vehicle_data[ATTR_STATE]:
                retval["lids"] = [
                    Lid(k, v)
                    for k, v in vehicle_data[ATTR_STATE]["doorsState"].items()
                    if k not in ["combinedState", "combinedSecurityState"] and v != LidState.INVALID
                ]
                retval["door_lock_state"] = LockState(
                    vehicle_data[ATTR_STATE]["doorsState"].get("combinedSecurityState", "UNKNOWN")
                )

            if "windowsState" in vehicle_data[ATTR_STATE]:
                retval["windows"] = [
                    Window(k, v)
                    for k, v in vehicle_data[ATTR_STATE]["windowsState"].items()
                    if k not in ["combinedState"] and v != LidState.INVALID
                ]

            if "roofState" in vehicle_data[ATTR_STATE]:
                retval["lids"].append(
                    Lid(
                        to_camel_case(vehicle_data[ATTR_STATE]["roofState"]["roofStateType"]),
                        vehicle_data[ATTR_STATE]["roofState"]["roofState"],
                    )
                )

        return retval

    @property
    def open_lids(self) -> List[Lid]:
        """Get all open lids of the car."""
        return [lid for lid in self.lids if not lid.is_closed]

    @property
    def all_lids_closed(self) -> bool:
        """Check if all lids are closed."""
        return len(self.open_lids) == 0

    @property
    def open_windows(self) -> List[Window]:
        """Get all open windows of the car."""
        return [lid for lid in self.windows if not lid.is_closed]

    @property
    def all_windows_closed(self) -> bool:
        """Check if all windows are closed."""
        return len(self.open_windows) == 0



"""Get the right url for the different countries."""
from base64 import b64decode
from typing import List

from bimmer_connected.const import APP_VERSIONS, OCP_APIM_KEYS, SERVER_URLS_MYBMW, USER_AGENTS, Regions


def valid_regions() -> List[str]:
    """Get list of valid regions as strings."""
    return [region.name.lower() for region in Regions]


def get_region_from_name(name: str) -> Regions:
    """Get a region for a string.

    This function is not case-sensitive.
    """
    for region in Regions:
        if name.lower() == region.name.lower():
            return region
    raise ValueError(f"Unknown region {name}. Valid regions are: {','.join(valid_regions())}")


def get_server_url(region: Regions) -> str:
    """Get the url of the server for the region."""
    return f"https://{SERVER_URLS_MYBMW[region]}"


def get_user_agent(region: Regions) -> str:
    """Get the Dart user agent for the region."""
    return USER_AGENTS[region]


def get_app_version(region: Regions) -> str:
    """Get the app version & build number for the region."""
    return APP_VERSIONS[region]


def get_ocp_apim_key(region: Regions) -> str:
    """Get the authorization for OAuth settings."""
    return b64decode(OCP_APIM_KEYS[region]).decode()



"""Generic API management."""

import logging
from collections import defaultdict, deque
from dataclasses import dataclass
from typing import Deque, Dict, Optional

import httpx

from bimmer_connected.api.authentication import MyBMWAuthentication
from bimmer_connected.api.regions import get_app_version, get_server_url, get_user_agent
from bimmer_connected.api.utils import anonymize_response, get_correlation_id, handle_httpstatuserror
from bimmer_connected.const import HTTPX_TIMEOUT, X_USER_AGENT, CarBrands
from bimmer_connected.models import AnonymizedResponse, GPSPosition

_LOGGER = logging.getLogger(__name__)

RESPONSE_STORE: Deque[AnonymizedResponse] = deque(maxlen=10)


@dataclass
class MyBMWClientConfiguration:
    """Stores global settings for MyBMWClient."""

    authentication: MyBMWAuthentication
    log_responses: Optional[bool] = False
    observer_position: Optional[GPSPosition] = None

    def set_log_responses(self, log_responses: bool) -> None:
        """Set if responses are logged and clear response store."""

        self.log_responses = log_responses
        RESPONSE_STORE.clear()


class MyBMWClient(httpx.AsyncClient):
    """Async HTTP client based on `httpx.AsyncClient` with automated OAuth token refresh."""

    def __init__(self, config: MyBMWClientConfiguration, *args, brand: Optional[CarBrands] = None, **kwargs):
        self.config = config

        # Add authentication
        kwargs["auth"] = self.config.authentication

        # Increase timeout
        kwargs["timeout"] = httpx.Timeout(HTTPX_TIMEOUT)

        # Set default values
        kwargs["base_url"] = kwargs.get("base_url") or get_server_url(config.authentication.region)
        kwargs["headers"] = kwargs.get("headers") or self.generate_default_header(brand)

        # Register event hooks
        kwargs["event_hooks"] = defaultdict(list, **kwargs.get("event_hooks", {}))

        # Event hook for logging content
        async def log_response(response: httpx.Response):
            await response.aread()
            RESPONSE_STORE.append(anonymize_response(response))

        if config.log_responses:
            kwargs["event_hooks"]["response"].append(log_response)

        # Event hook which calls raise_for_status on all requests
        async def raise_for_status_event_handler(response: httpx.Response):
            """Event handler that automatically raises HTTPStatusErrors when attached.

            Will only raise on 4xx/5xx errors but not 401/429 which are handled `self.auth`.
            """
            if response.is_error and response.status_code not in [401, 429]:
                try:
                    response.raise_for_status()
                except httpx.HTTPStatusError as ex:
                    await handle_httpstatuserror(ex, log_handler=_LOGGER)

        kwargs["event_hooks"]["response"].append(raise_for_status_event_handler)

        super().__init__(*args, **kwargs)

    def generate_default_header(self, brand: Optional[CarBrands] = None) -> Dict[str, str]:
        """Generate a header for HTTP requests to the server."""
        return {
            "accept": "application/json",
            "accept-language": "en",
            "user-agent": get_user_agent(self.config.authentication.region),
            "x-user-agent": X_USER_AGENT.format(
                brand=(brand or CarBrands.BMW).value,
                app_version=get_app_version(self.config.authentication.region),
                region=self.config.authentication.region.value,
            ),
            **get_correlation_id(),
            "bmw-units-preferences": "d=KM;v=L",
            "24-hour-format": "true",
        }



"""Access to a MyBMW account and all vehicles therein."""

import datetime
import json
import logging
from dataclasses import InitVar, dataclass, field
from typing import List, Optional

import httpx

from bimmer_connected.api.authentication import MyBMWAuthentication
from bimmer_connected.api.client import RESPONSE_STORE, MyBMWClient, MyBMWClientConfiguration
from bimmer_connected.api.regions import Regions
from bimmer_connected.const import (
    ATTR_CAPABILITIES,
    VEHICLE_CHARGING_DETAILS_URL,
    VEHICLE_STATE_URL,
    VEHICLES_URL,
    CarBrands,
)
from bimmer_connected.models import AnonymizedResponse, GPSPosition, MyBMWAPIError, MyBMWAuthError, MyBMWQuotaError
from bimmer_connected.vehicle import MyBMWVehicle

VALID_UNTIL_OFFSET = datetime.timedelta(seconds=10)

_LOGGER = logging.getLogger(__name__)


@dataclass
class MyBMWAccount:
    """Create a new connection to the MyBMW web service."""

    username: str
    """MyBMW user name (email) or 86-prefixed phone number (China only)."""

    password: InitVar[str]
    """MyBMW password."""

    region: Regions
    """Region of the account. See `api.Regions`."""

    config: MyBMWClientConfiguration = None  # type: ignore[assignment]
    """Optional. If provided, username/password/region are ignored."""

    log_responses: InitVar[bool] = False
    """Optional. If set, all responses from the server will be logged to this directory."""

    observer_position: InitVar[GPSPosition] = None
    """Optional. Required for getting a position on older cars."""

    use_metric_units: InitVar[Optional[bool]] = None
    """Deprecated. All returned values are metric units (km, l)."""

    vehicles: List[MyBMWVehicle] = field(default_factory=list, init=False)

    def __post_init__(self, password, log_responses, observer_position, use_metric_units):
        """Initialize the account."""

        if use_metric_units is not None:
            _LOGGER.warning(
                "The use_metric_units parameter is deprecated and will be removed in a future release. "
                "All values will be returned in metric units, as the parameter has no effect on the API."
            )

        if self.config is None:
            self.config = MyBMWClientConfiguration(
                MyBMWAuthentication(self.username, password, self.region),
                log_responses=log_responses,
                observer_position=observer_position,
            )

    async def _init_vehicles(self) -> None:
        """Initialize vehicles from BMW servers."""
        _LOGGER.debug("Getting vehicle list")

        fetched_at = datetime.datetime.now(datetime.timezone.utc)

        async with MyBMWClient(self.config) as client:
            vehicles_responses: List[httpx.Response] = [
                await client.get(
                    VEHICLES_URL,
                    headers={
                        **client.generate_default_header(brand),
                    },
                )
                for brand in CarBrands
            ]

            for response in vehicles_responses:
                for vehicle_base in response.json():
                    self.add_vehicle(vehicle_base, None, None, fetched_at)

    async def get_vehicles(self, force_init: bool = False) -> None:
        """Retrieve vehicle data from BMW servers."""
        _LOGGER.debug("Getting vehicle list")

        fetched_at = datetime.datetime.now(datetime.timezone.utc)

        if len(self.vehicles) == 0 or force_init:
            await self._init_vehicles()

        async with MyBMWClient(self.config) as client:
            error_count = 0
            for vehicle in self.vehicles:
                # Get the detailed vehicle state
                try:
                    state_response = await client.get(
                        VEHICLE_STATE_URL,
                        params={
                            "apptimezone": self.utcdiff,
                            "appDateTime": int(fetched_at.timestamp() * 1000),
                        },
                        headers={
                            **client.generate_default_header(vehicle.brand),
                            "bmw-vin": vehicle.vin,
                        },
                    )
                    vehicle_state = state_response.json()

                    # Get detailed charging settings if supported by vehicle
                    charging_settings = None
                    if vehicle_state[ATTR_CAPABILITIES].get("isChargingPlanSupported", False) or vehicle_state[
                        ATTR_CAPABILITIES
                    ].get("isChargingSettingsEnabled", False):
                        charging_settings_response = await client.get(
                            VEHICLE_CHARGING_DETAILS_URL,
                            params={
                                "fields": "charging-profile",
                                "has_charging_settings_capabilities": vehicle_state[ATTR_CAPABILITIES][
                                    "isChargingSettingsEnabled"
                                ],
                            },
                            headers={
                                **client.generate_default_header(vehicle.brand),
                                "bmw-current-date": fetched_at.isoformat(),
                                "bmw-vin": vehicle.vin,
                            },
                        )
                        charging_settings = charging_settings_response.json()

                    self.add_vehicle(vehicle.data, vehicle_state, charging_settings, fetched_at)
                except (MyBMWAPIError, json.JSONDecodeError) as ex:
                    # We don't want to fail completely if one vehicle fails, but we want to know about it
                    error_count += 1

                    # If it's a MyBMWQuotaError or MyBMWAuthError, we want to raise it
                    if isinstance(ex, (MyBMWQuotaError, MyBMWAuthError)):
                        raise ex

                    # Always log the error
                    _LOGGER.error("Unable to get details for vehicle %s - (%s) %s", vehicle.vin, type(ex).__name__, ex)

                    # If all vehicles fail, we want to raise an exception
                    if error_count == len(self.vehicles):
                        raise ex

    def add_vehicle(
        self,
        vehicle_base: dict,
        vehicle_state: Optional[dict],
        charging_settings: Optional[dict],
        fetched_at: Optional[datetime.datetime] = None,
    ) -> None:
        """Add or update a vehicle from the API responses."""

        existing_vehicle = self.get_vehicle(vehicle_base["vin"])

        # If vehicle already exists, just update it's state
        if existing_vehicle:
            existing_vehicle.update_state(vehicle_base, vehicle_state, charging_settings, fetched_at)
        else:
            self.vehicles.append(MyBMWVehicle(self, vehicle_base, vehicle_state, charging_settings, fetched_at))

    def get_vehicle(self, vin: str) -> Optional[MyBMWVehicle]:
        """Get vehicle with given VIN.

        The search is NOT case sensitive.
        :param vin: VIN of the vehicle you want to get.
        :return: Returns None if no vehicle is found.
        """
        for car in self.vehicles:
            if car.vin.upper() == vin.upper():
                return car
        return None

    def set_observer_position(self, latitude: float, longitude: float) -> None:
        """Set the position of the observer for all vehicles."""
        self.config.observer_position = GPSPosition(latitude=latitude, longitude=longitude)

    def set_refresh_token(self, refresh_token: str, gcid: Optional[str] = None) -> None:
        """Overwrite the current value of the MyBMW refresh token and GCID (if available)."""
        self.config.authentication.refresh_token = refresh_token
        self.config.authentication.gcid = gcid

    @staticmethod
    def get_stored_responses() -> List[AnonymizedResponse]:
        """Return responses stored if log_responses was set to True."""
        responses = list(RESPONSE_STORE)
        RESPONSE_STORE.clear()
        return responses

    @property
    def timezone(self):
        """Returns the current tzinfo."""
        return datetime.datetime.now().astimezone().tzinfo

    @property
    def utcdiff(self):
        """Returns the difference to UTC in minutes."""
        return round(self.timezone.utcoffset(datetime.datetime.now()).seconds / 60, 0)

    @property
    def refresh_token(self) -> Optional[str]:
        """Returns the current refresh_token."""
        return self.config.authentication.refresh_token

    @property
    def gcid(self) -> Optional[str]:
        """Returns the current GCID."""
        return self.config.authentication.gcid



"""Models the state of the vehicle tires."""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional

from bimmer_connected.const import ATTR_STATE
from bimmer_connected.models import VehicleDataBase


class TireState:
    """A tire of the vehicle."""

    def __init__(self, status: Dict, details: Optional[Dict] = None):
        self.current_pressure: Optional[int] = status.get("currentPressure")
        self.target_pressure: Optional[int] = status.get("targetPressure")
        self.season: Optional[int] = None
        self.manufacturing_week: Optional[datetime] = None

        if details:
            self.season = details.get("season")
            if details.get("manufacturingWeek"):
                iso_string = f"20{details['manufacturingWeek'] % 100}." f"{int(details['manufacturingWeek'] / 100)}.1"
                self.manufacturing_week = datetime.strptime(
                    iso_string,
                    "%G.%V.%u",
                )


@dataclass
class Tires(VehicleDataBase):
    """Provides an accessible version of `state.tireState`."""

    front_left: TireState
    """Front left tire info."""

    front_right: TireState
    """Front right tire info."""

    rear_left: TireState
    """Rear left tire info."""

    rear_right: TireState
    """Rear right tire info."""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Dict:
        """Parse tire status."""
        retval: Dict[str, Any] = {}

        if ATTR_STATE in vehicle_data:
            if "tireState" in vehicle_data[ATTR_STATE]:
                retval["front_left"] = TireState(**vehicle_data[ATTR_STATE]["tireState"]["frontLeft"])
                retval["front_right"] = TireState(**vehicle_data[ATTR_STATE]["tireState"]["frontRight"])
                retval["rear_left"] = TireState(**vehicle_data[ATTR_STATE]["tireState"]["rearLeft"])
                retval["rear_right"] = TireState(**vehicle_data[ATTR_STATE]["tireState"]["rearRight"])

        return retval



"""Models the charging profiles of a vehicle."""

import datetime
import logging
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from bimmer_connected.const import ATTR_CHARGING_SETTINGS, ATTR_STATE
from bimmer_connected.models import StrEnum, VehicleDataBase

_LOGGER = logging.getLogger(__name__)


class ChargingMode(StrEnum):
    """Charging mode of electric vehicle."""

    IMMEDIATE_CHARGING = "IMMEDIATE_CHARGING"
    DELAYED_CHARGING = "DELAYED_CHARGING"
    UNKNOWN = "UNKNOWN"


MAP_CHARGING_MODE_TO_REMOTE_SERVICE = {
    ChargingMode.IMMEDIATE_CHARGING: "CHARGING_IMMEDIATELY",
    ChargingMode.DELAYED_CHARGING: "TIME_SLOT",
}


class ChargingPreferences(StrEnum):
    """Charging preferences of electric vehicle."""

    NO_PRESELECTION = "NO_PRESELECTION"
    CHARGING_WINDOW = "CHARGING_WINDOW"
    UNKNOWN = "UNKNOWN"


class TimerTypes(StrEnum):
    """Different timer types."""

    WEEKLY_PLANNER = "WEEKLY_PLANNER"
    TWO_TIMES_TIMER = "TWO_TIMES_TIMER"
    UNKNOWN = "UNKNOWN"


MAP_TIMER_TYPES_TO_REMOTE_SERVICE = {
    TimerTypes.WEEKLY_PLANNER: "WEEKLY_DEPARTURE_TIMER",
    TimerTypes.TWO_TIMES_TIMER: "TWO_DEPARTURE_TIMER",
}


class ChargingWindow:
    """A charging window."""

    def __init__(self, window_dict: dict):
        self._window_dict = window_dict

    @property
    def start_time(self) -> datetime.time:
        """Start of the charging window."""
        if "start" not in self._window_dict:
            return datetime.time(0, 0)
        return datetime.time(int(self._window_dict["start"]["hour"]), int(self._window_dict["start"]["minute"]))

    @property
    def end_time(self) -> datetime.time:
        """End of the charging window."""
        if "end" not in self._window_dict:
            return datetime.time(0, 0)
        return datetime.time(int(self._window_dict["end"]["hour"]), int(self._window_dict["end"]["minute"]))


class DepartureTimer:
    """A departure timer."""

    def __init__(self, timer_dict: dict):
        self._timer_dict: Dict = timer_dict

    @property
    def timer_id(self) -> Optional[int]:
        """ID of this timer."""
        return self._timer_dict.get("id")

    @property
    def start_time(self) -> Optional[datetime.time]:
        """Deperture time for this timer."""
        if "timeStamp" not in self._timer_dict:
            return None
        return datetime.time(int(self._timer_dict["timeStamp"]["hour"]), int(self._timer_dict["timeStamp"]["minute"]))

    @property
    def action(self) -> Optional[str]:
        """What does the timer do."""
        return self._timer_dict.get("action")

    @property
    def weekdays(self) -> List[str]:
        """Active weekdays for this timer."""
        return self._timer_dict.get("timerWeekDays")  # type: ignore[return-value]


@dataclass
class ChargingProfile(VehicleDataBase):
    """Models the charging profile of a vehicle."""

    is_pre_entry_climatization_enabled: bool
    """Get status of pre-entry climatization."""

    timer_type: TimerTypes
    """Returns the current timer plan type."""

    departure_times: List[DepartureTimer]
    """List of timers."""

    preferred_charging_window: ChargingWindow
    """Returns the preferred charging window."""

    charging_preferences: ChargingPreferences
    """Returns the preferred charging preferences."""

    charging_mode: ChargingMode
    """Returns the preferred charging mode."""

    ac_current_limit: Optional[int] = None
    """Returns the ac current limit."""

    ac_available_limits: Optional[list] = None
    """Available AC limits to be selected."""

    charging_preferences_service_pack: Optional[str] = None
    """Service Pack required for remote service format."""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Dict:
        """Parse charging data."""
        retval: Dict[str, Any] = {}

        if ATTR_STATE in vehicle_data and "chargingProfile" in vehicle_data[ATTR_STATE]:
            charging_profile = vehicle_data[ATTR_STATE]["chargingProfile"]

            retval["is_pre_entry_climatization_enabled"] = bool(charging_profile.get("climatisationOn", False))
            retval["departure_times"] = [DepartureTimer(t) for t in charging_profile.get("departureTimes", [])]
            retval["preferred_charging_window"] = ChargingWindow(charging_profile.get("reductionOfChargeCurrent", {}))
            retval["timer_type"] = TimerTypes(charging_profile.get("chargingControlType", "UNKNOWN"))
            retval["charging_preferences"] = ChargingPreferences(charging_profile.get("chargingPreference", "UNKNOWN"))
            retval["charging_mode"] = ChargingMode(charging_profile.get("chargingMode", "UNKNOWN"))
            if "acCurrentLimit" in charging_profile["chargingSettings"]:
                retval["ac_current_limit"] = charging_profile["chargingSettings"]["acCurrentLimit"]

        if ATTR_CHARGING_SETTINGS in vehicle_data:
            charging_settings = vehicle_data[ATTR_CHARGING_SETTINGS]
            if "servicePack" in charging_settings:
                retval["charging_preferences_service_pack"] = charging_settings["servicePack"]
            if (
                "chargingSettingsDetail" in charging_settings
                and "acLimit" in charging_settings["chargingSettingsDetail"]
            ):
                retval["ac_available_limits"] = charging_settings["chargingSettingsDetail"]["acLimit"]["values"]

        return retval

    def format_for_remote_service(self) -> dict:
        """Format current charging profile as base to be sent to remote service."""

        return {
            "chargingMode": {
                "chargingPreference": self.charging_preferences.value,
                "endTimeSlot": self._format_time(self.preferred_charging_window.end_time),
                "startTimeSlot": self._format_time(self.preferred_charging_window.start_time),
                "type": MAP_CHARGING_MODE_TO_REMOTE_SERVICE[self.charging_mode],
                "timerChange": "NO_CHANGE",
            },
            "departureTimer": {
                "type": MAP_TIMER_TYPES_TO_REMOTE_SERVICE[self.timer_type],
                "weeklyTimers": [
                    {
                        "daysOfTheWeek": t.weekdays,
                        "id": t.timer_id,
                        "time": self._format_time(t.start_time),
                        "timerAction": t.action,
                    }
                    for t in self.departure_times
                ],
            },
            "isPreconditionForDepartureActive": self.is_pre_entry_climatization_enabled,
            "servicePack": self.charging_preferences_service_pack,
        }

    @staticmethod
    def _format_time(time: Optional[datetime.time] = None) -> str:
        if not time:
            return "0001-01-01T00:00:00.000"
        return time.strftime("0001-01-01T%H:%M:00.000")



"""Tests for MyBMWVehicle."""
import pytest
import respx

from bimmer_connected.const import ATTR_ATTRIBUTES, ATTR_STATE, CarBrands
from bimmer_connected.models import GPSPosition, StrEnum, VehicleDataBase
from bimmer_connected.vehicle import VehicleViewDirection
from bimmer_connected.vehicle.const import DriveTrainType
from bimmer_connected.vehicle.reports import CheckControlMessageReport

from . import (
    VIN_F31,
    VIN_G01,
    VIN_G20,
    VIN_G26,
    VIN_G70,
    VIN_I01_NOREX,
    VIN_I01_REX,
    VIN_I20,
    get_deprecation_warning_count,
)
from .conftest import prepare_account_with_vehicles

ATTRIBUTE_MAPPING = {
    "remainingFuel": "remaining_fuel",
    "position": "gps_position",
    "cbsData": "condition_based_services",
    "checkControlMessages": "check_control_messages",
    "doorLockState": "door_lock_state",
    "updateReason": "last_update_reason",
    "chargingLevelHv": "charging_level_hv",
    "chargingStatus": "charging_status",
    "maxRangeElectric": "max_range_electric",
    "remainingRangeElectric": "remaining_range_electric",
    "parkingLight": "parking_lights",
    "remainingRangeFuel": "remaining_range_fuel",
    "updateTime": "timestamp",
    "chargingTimeRemaining": "charging_time_remaining",
}


@pytest.mark.asyncio
async def test_drive_train(caplog, bmw_fixture: respx.Router):
    """Tests around drive_train attribute."""
    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_F31)
    assert DriveTrainType.COMBUSTION == vehicle.drive_train

    vehicle = account.get_vehicle(VIN_G01)
    assert DriveTrainType.PLUGIN_HYBRID == vehicle.drive_train

    vehicle = account.get_vehicle(VIN_G26)
    assert DriveTrainType.ELECTRIC == vehicle.drive_train

    vehicle = account.get_vehicle(VIN_I01_NOREX)
    assert DriveTrainType.ELECTRIC == vehicle.drive_train

    vehicle = account.get_vehicle(VIN_I01_REX)
    assert DriveTrainType.ELECTRIC_WITH_RANGE_EXTENDER == vehicle.drive_train

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_parsing_attributes(caplog, bmw_fixture: respx.Router):
    """Test parsing different attributes of the vehicle."""
    account = await prepare_account_with_vehicles()

    for vehicle in account.vehicles:
        print(vehicle.name)
        assert vehicle.drive_train is not None
        assert vehicle.name is not None
        assert isinstance(vehicle.brand, CarBrands)
        assert vehicle.has_combustion_drivetrain is not None
        assert vehicle.has_electric_drivetrain is not None
        assert vehicle.drive_train_attributes is not None
        assert vehicle.is_charging_plan_supported is not None

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_drive_train_attributes(caplog, bmw_fixture: respx.Router):
    """Test parsing different attributes of the vehicle."""
    account = await prepare_account_with_vehicles()

    vehicle_drivetrains = {
        VIN_F31: (True, False),
        VIN_G01: (True, True),
        VIN_G20: (True, False),
        VIN_G26: (False, True),
        VIN_G70: (False, True),
        VIN_I01_NOREX: (False, True),
        VIN_I01_REX: (True, True),
        VIN_I20: (False, True),
    }

    for vehicle in account.vehicles:
        assert vehicle_drivetrains[vehicle.vin][0] == vehicle.has_combustion_drivetrain
        assert vehicle_drivetrains[vehicle.vin][1] == vehicle.has_electric_drivetrain

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_parsing_of_lsc_type(caplog, bmw_fixture: respx.Router):
    """Test parsing the lsc type field."""
    account = await prepare_account_with_vehicles()

    for vehicle in account.vehicles:
        assert vehicle.lsc_type is not None

    assert len(get_deprecation_warning_count(caplog)) == 0


def test_car_brand(caplog, bmw_fixture: respx.Router):
    """Test CarBrand enum."""
    assert CarBrands("BMW") == CarBrands("bmw")

    with pytest.raises(ValueError):
        CarBrands("Audi")

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_get_is_tracking_enabled(caplog, bmw_fixture: respx.Router):
    """Test setting observer position."""
    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_I01_REX)
    assert vehicle.is_vehicle_tracking_enabled is False

    vehicle = account.get_vehicle(VIN_F31)
    assert vehicle.is_vehicle_tracking_enabled is True

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_available_attributes(caplog, bmw_fixture: respx.Router):
    """Check that available_attributes returns exactly the arguments we have in our test data."""
    account = await prepare_account_with_vehicles()

    vehicle = account.get_vehicle(VIN_F31)
    assert ["gps_position", "vin"] == vehicle.available_attributes

    vehicle = account.get_vehicle(VIN_G01)
    assert [
        "gps_position",
        "vin",
        "remaining_range_total",
        "mileage",
        "charging_time_remaining",
        "charging_start_time",
        "charging_end_time",
        "charging_time_label",
        "charging_status",
        "connection_status",
        "remaining_battery_percent",
        "remaining_range_electric",
        "last_charging_end_result",
        "ac_current_limit",
        "charging_target",
        "charging_mode",
        "charging_preferences",
        "is_pre_entry_climatization_enabled",
        "remaining_fuel",
        "remaining_range_fuel",
        "remaining_fuel_percent",
        "condition_based_services",
        "check_control_messages",
        "door_lock_state",
        "timestamp",
        "lids",
        "windows",
    ] == vehicle.available_attributes

    vehicle = account.get_vehicle(VIN_G26)
    assert [
        "gps_position",
        "vin",
        "remaining_range_total",
        "mileage",
        "charging_time_remaining",
        "charging_start_time",
        "charging_end_time",
        "charging_time_label",
        "charging_status",
        "connection_status",
        "remaining_battery_percent",
        "remaining_range_electric",
        "last_charging_end_result",
        "ac_current_limit",
        "charging_target",
        "charging_mode",
        "charging_preferences",
        "is_pre_entry_climatization_enabled",
        "condition_based_services",
        "check_control_messages",
        "door_lock_state",
        "timestamp",
        "lids",
        "windows",
    ] == vehicle.available_attributes

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_vehicle_image(caplog, bmw_fixture: respx.Router):
    """Test vehicle image request."""
    vehicle = (await prepare_account_with_vehicles()).get_vehicle(VIN_G01)

    bmw_fixture.get(
        path__regex=r"(.*)/eadrax-ics/v3/presentation/vehicles/\w*/images",
        params={"carView": "FrontView"},
        headers={"accept": "image/png"},
    ).respond(200, content="png_image")
    assert b"png_image" == await vehicle.get_vehicle_image(VehicleViewDirection.FRONT)

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_no_timestamp(bmw_fixture: respx.Router):
    """Test no timestamp available."""
    vehicle = (await prepare_account_with_vehicles()).get_vehicle(VIN_F31)
    vehicle.data[ATTR_STATE].pop("lastFetched")
    vehicle.data[ATTR_ATTRIBUTES].pop("lastFetched")

    assert vehicle.timestamp is None


def test_strenum(caplog):
    """Tests StrEnum."""

    class TestEnum(StrEnum):
        """Test StrEnum."""

        HELLO = "HELLO"

    assert TestEnum("hello") == TestEnum.HELLO
    assert TestEnum("HELLO") == TestEnum.HELLO

    with pytest.raises(ValueError):
        TestEnum("WORLD")

    class TestEnumUnkown(StrEnum):
        """Test StrEnum with UNKNOWN value."""

        HELLO = "HELLO"
        UNKNOWN = "UNKNOWN"

    assert TestEnumUnkown("hello") == TestEnumUnkown.HELLO
    assert TestEnumUnkown("HELLO") == TestEnumUnkown.HELLO

    assert len([r for r in caplog.records if r.levelname == "WARNING"]) == 0
    assert TestEnumUnkown("WORLD") == TestEnumUnkown.UNKNOWN
    assert len([r for r in caplog.records if r.levelname == "WARNING"]) == 1


def test_vehiclebasedata():
    """Tests VehicleBaseData."""
    with pytest.raises(NotImplementedError):
        VehicleDataBase._parse_vehicle_data({})

    # CheckControlMessageReport does not override parent methods from_vehicle_data()
    ccmr = CheckControlMessageReport.from_vehicle_data(
        {"state": {"checkControlMessages": [{"severity": "LOW", "type": "ENGINE_OIL"}]}}
    )
    assert len(ccmr.messages) == 1
    assert ccmr.has_check_control_messages is False


def test_gpsposition():
    """Tests around GPSPosition."""
    pos = GPSPosition(1.0, 2.0)
    assert pos == GPSPosition(1, 2)
    assert pos == {"latitude": 1.0, "longitude": 2.0}
    assert pos == (1, 2)
    assert pos != "(1, 2)"
    assert pos[0] == 1


@pytest.mark.asyncio
async def test_headunit_data(caplog, bmw_fixture: respx.Router):
    """Test if the parsing of headunit is working."""

    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_I20).headunit

    assert status.idrive_version == "ID8"
    assert status.headunit_type == "MGU"
    assert status.software_version == "07/2021.00"

    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_F31).headunit

    assert status.idrive_version == "ID4"
    assert status.headunit_type == "NBT"
    assert status.software_version == "11/2013.02"

    assert len(get_deprecation_warning_count(caplog)) == 0



"""Test for remote_services."""
from unittest import mock
from uuid import uuid4

import httpx
import pytest
import respx
import time_machine

from bimmer_connected.api.client import MyBMWClient
from bimmer_connected.models import MyBMWAPIError, MyBMWRemoteServiceError, PointOfInterest
from bimmer_connected.vehicle import remote_services
from bimmer_connected.vehicle.charging_profile import ChargingMode
from bimmer_connected.vehicle.climate import ClimateActivityState
from bimmer_connected.vehicle.doors_windows import LockState
from bimmer_connected.vehicle.fuel_and_battery import ChargingState
from bimmer_connected.vehicle.remote_services import ExecutionState, RemoteServiceStatus

from . import (
    REMOTE_SERVICE_RESPONSE_DELIVERED,
    REMOTE_SERVICE_RESPONSE_ERROR,
    REMOTE_SERVICE_RESPONSE_EXECUTED,
    REMOTE_SERVICE_RESPONSE_PENDING,
    VIN_F31,
    VIN_G01,
    VIN_G26,
    VIN_I01_NOREX,
    VIN_I20,
    load_response,
)
from .common import (
    CHARGING_SETTINGS,
    POI_DATA,
)
from .conftest import prepare_account_with_vehicles

remote_services._POLLING_CYCLE = 0


def test_states():
    """Test parsing the different response types."""
    rss = RemoteServiceStatus(load_response(REMOTE_SERVICE_RESPONSE_PENDING))
    assert ExecutionState.PENDING == rss.state

    rss = RemoteServiceStatus(load_response(REMOTE_SERVICE_RESPONSE_DELIVERED))
    assert ExecutionState.DELIVERED == rss.state

    rss = RemoteServiceStatus(load_response(REMOTE_SERVICE_RESPONSE_EXECUTED))
    assert ExecutionState.EXECUTED == rss.state


ALL_SERVICES = {
    "LIGHT_FLASH": {"call": "trigger_remote_light_flash", "refresh": False},
    "DOOR_LOCK": {"call": "trigger_remote_door_lock", "refresh": True},
    "DOOR_UNLOCK": {"call": "trigger_remote_door_unlock", "refresh": True},
    "CLIMATE_NOW": {"call": "trigger_remote_air_conditioning", "refresh": True},
    "CLIMATE_STOP": {"call": "trigger_remote_air_conditioning_stop", "refresh": True},
    "VEHICLE_FINDER": {"call": "trigger_remote_vehicle_finder", "refresh": False},
    "HORN_BLOW": {"call": "trigger_remote_horn", "refresh": False},
    "SEND_POI": {"call": "trigger_send_poi", "refresh": False, "args": [POI_DATA]},
    "CHARGE_START": {"call": "trigger_charge_start", "refresh": True},
    "CHARGE_STOP": {"call": "trigger_charge_stop", "refresh": True},
    "CHARGING_SETTINGS": {"call": "trigger_charging_settings_update", "refresh": True, "kwargs": CHARGING_SETTINGS},
}


@pytest.mark.asyncio
@pytest.mark.filterwarnings("ignore:coroutine 'AsyncMockMixin._execute_mock_call' was never awaited:RuntimeWarning")
async def test_trigger_remote_services(bmw_fixture: respx.Router):
    """Test executing a remote light flash."""

    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_I20)

    for service in ALL_SERVICES.values():
        with mock.patch(
            "bimmer_connected.account.MyBMWAccount.get_vehicles", new_callable=mock.AsyncMock
        ) as mock_listener:
            mock_listener.reset_mock()

            response = await getattr(vehicle.remote_services, service["call"])(  # type: ignore[call-overload]
                *service.get("args", []), **service.get("kwargs", {})
            )
            assert ExecutionState.EXECUTED == response.state

            if service["refresh"]:
                mock_listener.assert_called_once_with()
            else:
                mock_listener.assert_not_called()


@pytest.mark.asyncio
async def test_get_remote_service_status(bmw_fixture: respx.Router):
    """Test get_remove_service_status method."""

    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_G26)
    client = MyBMWClient(account.config)

    bmw_fixture.post("/eadrax-vrccs/v3/presentation/remote-commands/eventStatus", params={"eventId": mock.ANY}).mock(
        side_effect=[
            httpx.Response(500),
            httpx.Response(200, text="You can't parse this..."),
            httpx.Response(200, json=load_response(REMOTE_SERVICE_RESPONSE_ERROR)),
        ],
    )

    with pytest.raises(MyBMWAPIError):
        await vehicle.remote_services._block_until_done(client, uuid4())
    with pytest.raises(ValueError):
        await vehicle.remote_services._block_until_done(client, uuid4())
    with pytest.raises(MyBMWRemoteServiceError):
        await vehicle.remote_services._block_until_done(client, uuid4())


@pytest.mark.asyncio
async def test_set_lock_result(bmw_fixture: respx.Router):
    """Test locking/unlocking a car."""

    account = await prepare_account_with_vehicles()

    vehicle = account.get_vehicle(VIN_I01_NOREX)
    # check current state, unlock vehicle, check changed state
    assert vehicle.doors_and_windows.door_lock_state == LockState.UNLOCKED
    await vehicle.remote_services.trigger_remote_door_lock()
    assert vehicle.doors_and_windows.door_lock_state == LockState.LOCKED

    # now lock vehicle again, check changed state
    await vehicle.remote_services.trigger_remote_door_unlock()
    assert vehicle.doors_and_windows.door_lock_state == LockState.UNLOCKED


@pytest.mark.asyncio
async def test_set_climate_result(bmw_fixture: respx.Router):
    """Test starting/stopping climatization."""

    account = await prepare_account_with_vehicles()

    vehicle = account.get_vehicle(VIN_G01)
    # check current state, unlock vehicle, check changed state
    assert vehicle.climate.activity == ClimateActivityState.STANDBY
    await vehicle.remote_services.trigger_remote_air_conditioning()
    assert vehicle.climate.activity in [ClimateActivityState.COOLING, ClimateActivityState.HEATING]

    # now lock vehicle again, check changed state
    await vehicle.remote_services.trigger_remote_air_conditioning_stop()
    assert vehicle.climate.activity == ClimateActivityState.STANDBY


@pytest.mark.asyncio
async def test_charging_start_stop(bmw_fixture: respx.Router):
    """Test starting/stopping climatization."""

    account = await prepare_account_with_vehicles()

    vehicle = account.get_vehicle(VIN_I20)

    # check current state, unlock vehicle, check changed state
    assert vehicle.fuel_and_battery.charging_status == ChargingState.CHARGING
    await vehicle.remote_services.trigger_charge_stop()
    assert vehicle.fuel_and_battery.charging_status == ChargingState.PLUGGED_IN

    # now lock vehicle again, check changed state
    await vehicle.remote_services.trigger_charge_start()
    assert vehicle.fuel_and_battery.charging_status == ChargingState.CHARGING


@pytest.mark.asyncio
async def test_set_charging_settings(bmw_fixture: respx.Router):
    """Test setting the charging settings on a car."""

    account = await prepare_account_with_vehicles()

    # Errors on old electric vehicles, combustion engines and PHEV
    for vin in [VIN_I01_NOREX, VIN_F31, VIN_G01]:
        vehicle = account.get_vehicle(vin)
        with pytest.raises(ValueError):
            await vehicle.remote_services.trigger_charging_settings_update(target_soc=80)
        with pytest.raises(ValueError):
            await vehicle.remote_services.trigger_charging_settings_update(ac_limit=16)

    # This should work
    vehicle = account.get_vehicle(VIN_G26)
    # Test current state
    assert vehicle.charging_profile.ac_current_limit == 16
    assert vehicle.fuel_and_battery.charging_target == 80
    # Update settings
    await vehicle.remote_services.trigger_charging_settings_update(target_soc=75, ac_limit=12)
    # Test changed state
    assert vehicle.charging_profile.ac_current_limit == 12
    assert vehicle.fuel_and_battery.charging_target == 75

    # But these are not allowed
    with pytest.raises(ValueError):
        await vehicle.remote_services.trigger_charging_settings_update(target_soc=19)
    with pytest.raises(ValueError):
        await vehicle.remote_services.trigger_charging_settings_update(target_soc=21)
    with pytest.raises(ValueError):
        await vehicle.remote_services.trigger_charging_settings_update(target_soc=101)
    with pytest.raises(ValueError):
        await vehicle.remote_services.trigger_charging_settings_update(target_soc="asdf")
    with pytest.raises(ValueError):
        await vehicle.remote_services.trigger_charging_settings_update(ac_limit=17)
    with pytest.raises(ValueError):
        await vehicle.remote_services.trigger_charging_settings_update(ac_limit="asdf")


@pytest.mark.asyncio
async def test_set_charging_profile(bmw_fixture: respx.Router):
    """Test setting the charging profile on a car."""

    account = await prepare_account_with_vehicles()

    # Errors on combustion engines
    vehicle = account.get_vehicle(VIN_F31)
    with pytest.raises(ValueError):
        await vehicle.remote_services.trigger_charging_profile_update(precondition_climate=True)

    # This shouldn't fail even on older EV
    vehicle = account.get_vehicle(VIN_I01_NOREX)
    # check current state
    assert vehicle.charging_profile.charging_mode == ChargingMode.IMMEDIATE_CHARGING
    assert vehicle.charging_profile.is_pre_entry_climatization_enabled is True

    # update two settings
    await vehicle.remote_services.trigger_charging_profile_update(
        charging_mode=ChargingMode.DELAYED_CHARGING, precondition_climate=False
    )
    assert vehicle.charging_profile.charging_mode == ChargingMode.DELAYED_CHARGING
    assert vehicle.charging_profile.is_pre_entry_climatization_enabled is False

    # change back only charging mode
    await vehicle.remote_services.trigger_charging_profile_update(charging_mode=ChargingMode.IMMEDIATE_CHARGING)
    assert vehicle.charging_profile.charging_mode == ChargingMode.IMMEDIATE_CHARGING

    # change back only climatization
    await vehicle.remote_services.trigger_charging_profile_update(precondition_climate=True)
    assert vehicle.charging_profile.is_pre_entry_climatization_enabled is True


@pytest.mark.asyncio
async def test_vehicles_without_enabled_services(bmw_fixture: respx.Router):
    """Test setting the charging profile on a car."""

    account = await prepare_account_with_vehicles()

    # Errors on combustion engines
    vehicle = account.get_vehicle(VIN_F31)

    vehicle.update_state(vehicle.data, {"capabilities": {}})

    for service in ALL_SERVICES.values():
        with pytest.raises(ValueError):
            await getattr(vehicle.remote_services, service["call"])(  # type: ignore[call-overload]
                *service.get("args", []), **service.get("kwargs", {})
            )


@pytest.mark.asyncio
async def test_trigger_charge_start_stop_warnings(caplog, bmw_fixture: respx.Router):
    """Test if warnings are produced correctly with the charge start/stop services."""

    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_I20)

    fixture_not_connected = {
        **vehicle.data["state"]["electricChargingState"],
        "chargingStatus": "INVALID",
        "isChargerConnected": False,
    }
    vehicle.update_state(vehicle.data, {"state": {"electricChargingState": fixture_not_connected}})

    result = await vehicle.remote_services.trigger_charge_start()
    assert result.state == ExecutionState.IGNORED
    assert len([r for r in caplog.records if r.levelname == "WARNING" and "Charger not connected" in r.message]) == 1
    caplog.clear()

    result = await vehicle.remote_services.trigger_charge_stop()
    assert result.state == ExecutionState.IGNORED
    assert len([r for r in caplog.records if r.levelname == "WARNING" and "Charger not connected" in r.message]) == 1
    caplog.clear()

    fixture_connected_not_charging = {
        **vehicle.data["state"]["electricChargingState"],
        "chargingStatus": "WAITING_FOR_CHARGING",
        "isChargerConnected": True,
    }
    vehicle.update_state(vehicle.data, {"state": {"electricChargingState": fixture_connected_not_charging}})

    result = await vehicle.remote_services.trigger_charge_stop()
    assert result.state == ExecutionState.IGNORED
    assert len([r for r in caplog.records if r.levelname == "WARNING" and "Vehicle not charging" in r.message]) == 1
    caplog.clear()


@pytest.mark.asyncio
async def test_get_remote_position(bmw_fixture: respx.Router):
    """Test getting position from remote service."""

    account = await prepare_account_with_vehicles()
    account.set_observer_position(1.0, 0.0)
    vehicle = account.get_vehicle(VIN_G26)
    location = vehicle.vehicle_location

    # Check original position
    assert (48.177334, 11.556274) == location.location
    assert 180 == location.heading

    # Check updated position
    await vehicle.remote_services.trigger_remote_vehicle_finder()
    assert (123.456, 34.5678) == location.location
    assert 121 == location.heading

    # Position should still be from vehicle finder after status update
    await account.get_vehicles()
    assert (123.456, 34.5678) == location.location
    assert 121 == location.heading


@pytest.mark.asyncio
async def test_get_remote_position_fail_without_observer(caplog, bmw_fixture: respx.Router):
    """Test getting position from remote service."""

    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_G26)

    await vehicle.remote_services.trigger_remote_vehicle_finder()
    errors = [
        r
        for r in caplog.records
        if r.levelname == "ERROR"
        and "Unknown position: Set observer position to retrieve vehicle coordinates" in r.message
    ]
    assert len(errors) == 1


@pytest.mark.asyncio
async def test_fail_with_timeout(bmw_fixture: respx.Router):
    """Test failing after timeout was reached."""
    remote_services._POLLING_CYCLE = 1
    remote_services._POLLING_TIMEOUT = 2

    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_G26)

    with pytest.raises(MyBMWRemoteServiceError):
        await vehicle.remote_services.trigger_remote_light_flash()


@time_machine.travel("2020-01-01", tick=False)
@pytest.mark.asyncio
async def test_get_remote_position_too_old(bmw_fixture: respx.Router):
    """Test remote service position being ignored as vehicle status is newer."""

    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_G26)
    location = vehicle.vehicle_location

    await vehicle.remote_services.trigger_remote_vehicle_finder()

    assert (48.177334, 11.556274) == location.location
    assert 180 == location.heading


@pytest.mark.asyncio
async def test_poi(bmw_fixture: respx.Router):
    """Test get_remove_service_status method."""

    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_G26)

    await vehicle.remote_services.trigger_send_poi({"lat": 12.34, "lon": 12.34})

    with pytest.raises(TypeError):
        await vehicle.remote_services.trigger_send_poi({"lat": 12.34})


def test_poi_parsing():
    """Test correct parsing of PointOfInterest."""

    poi_data = PointOfInterest(**POI_DATA)

    # Check parsing of attributes required by API
    assert poi_data.coordinates.latitude == POI_DATA["lat"]
    assert poi_data.coordinates.longitude == POI_DATA["lon"]
    assert poi_data.name == POI_DATA["name"]
    assert poi_data.formattedAddress == f"{POI_DATA['street']}, {POI_DATA['postal_code']}, {POI_DATA['city']}"

    # Check the default attributes
    poi_data = PointOfInterest(lat=POI_DATA["lat"], lon=POI_DATA["lon"])
    assert poi_data.coordinates.latitude == POI_DATA["lat"]
    assert poi_data.coordinates.longitude == POI_DATA["lon"]
    assert poi_data.name == "Sent with ♥ by bimmer_connected"
    assert poi_data.formattedAddress == "Coordinates only"

    # Check the default attributes with formatted address
    poi_data = PointOfInterest(lat=POI_DATA["lat"], lon=POI_DATA["lon"], formattedAddress="Somewhere over rainbow")
    assert poi_data.coordinates.latitude == POI_DATA["lat"]
    assert poi_data.coordinates.longitude == POI_DATA["lon"]
    assert poi_data.name == "Sent with ♥ by bimmer_connected"
    assert poi_data.formattedAddress == "Somewhere over rainbow"



"""Generals models used for bimmer_connected."""

import logging
from dataclasses import InitVar, dataclass, field
from enum import Enum
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

from bimmer_connected.const import DEFAULT_POI_NAME

_LOGGER = logging.getLogger(__name__)


class StrEnum(str, Enum):
    """A string enumeration of type `(str, Enum)`. All members are compared via `upper()`. Defaults to UNKNOWN."""

    @classmethod
    def _missing_(cls, value):
        has_unknown = False
        for member in cls:
            if member.value.upper() == "UNKNOWN":
                has_unknown = True
            if member.value.upper() == value.upper():
                return member
        if has_unknown:
            _LOGGER.warning("'%s' is not a valid '%s'", value, cls.__name__)
            return getattr(cls, "UNKNOWN")
        raise ValueError(f"'{value}' is not a valid {cls.__name__}")


@dataclass
class VehicleDataBase:
    """A base class for parsing and storing complex vehicle data."""

    @classmethod
    def from_vehicle_data(cls, vehicle_data: Dict):
        """Create the class based on vehicle data from API."""
        parsed = cls._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return cls(**parsed)
        return None

    def update_from_vehicle_data(self, vehicle_data: Dict):
        """Update the attributes based on vehicle data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        parsed.update(self._update_after_parse(parsed))
        if len(parsed) > 0:
            self.__dict__.update(parsed)

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parse desired attributes out of vehicle data from API."""
        raise NotImplementedError()

    def _update_after_parse(self, parsed: Dict) -> Dict:
        """Update parsed vehicle data with attributes stored in class if needed."""
        return parsed


@dataclass
class GPSPosition:
    """GPS coordinates."""

    latitude: Optional[float]
    longitude: Optional[float]

    def __post_init__(self):
        if len([v for v in self.__dict__.values() if v is None]) not in [0, len(self.__dataclass_fields__)]:
            raise TypeError("Either none or all arguments must be 'None'.")

        for field_name in self.__dataclass_fields__:
            value = getattr(self, field_name)
            if value is not None and not isinstance(value, (float, int)):
                raise TypeError(f"'{field_name}' not of type '{Optional[Union[float, int]]}'")

    def __iter__(self):
        yield from self.__dict__.values()

    def __getitem__(self, key):
        return tuple(self.__dict__.values())[key]

    def __eq__(self, other):
        if isinstance(other, Tuple):
            return tuple(self.__iter__()) == other
        if hasattr(self, "__dict__") and hasattr(other, "__dict__"):
            return self.__dict__ == other.__dict__
        if hasattr(self, "__dict__") and isinstance(other, Dict):
            return self.__dict__ == other
        return False


@dataclass
class PointOfInterestAddress:
    """Address data of a PointOfInterest."""

    street: Optional[str] = None
    postalCode: Optional[str] = None
    city: Optional[str] = None
    country: Optional[str] = None

    # The following attributes are not by us but available in the API
    banchi: Optional[str] = None
    chome: Optional[str] = None
    countryCode: Optional[str] = None
    district: Optional[str] = None
    go: Optional[str] = None
    houseNumber: Optional[str] = None
    region: Optional[str] = None
    regionCode: Optional[str] = None
    settlement: Optional[str] = None


@dataclass
class PointOfInterest:
    """A Point of Interest to be sent to the car."""

    lat: InitVar[float]
    lon: InitVar[float]
    name: Optional[str] = DEFAULT_POI_NAME
    street: InitVar[str] = None
    postal_code: InitVar[str] = None
    city: InitVar[str] = None
    country: InitVar[str] = None

    coordinates: GPSPosition = field(init=False)
    locationAddress: Optional[PointOfInterestAddress] = field(init=False)
    # The following attributes are not by us but required in the API
    formattedAddress: Optional[str] = None
    entryPoints: List = field(init=False, default_factory=list)

    # The following attributes are not by us but available in the API
    address: Optional[str] = None
    baseCategoryId: Optional[str] = None
    phoneNumber: Optional[str] = None
    provider: Optional[str] = None
    providerId: Optional[str] = None
    providerPoiId: str = ""
    sourceType: Optional[str] = None
    type: Optional[str] = None
    vehicleCategoryId: Optional[str] = None

    def __post_init__(self, lat, lon, street, postal_code, city, country):
        self.coordinates = GPSPosition(lat, lon)

        self.locationAddress = PointOfInterestAddress(street, postal_code, city, country)

        if not self.formattedAddress:
            self.formattedAddress = ", ".join([i for i in [street, postal_code, city] if i]) or "Coordinates only"


class ValueWithUnit(NamedTuple):
    """A value with a corresponding unit."""

    value: Optional[Union[int, float]]
    unit: Optional[str]


@dataclass
class AnonymizedResponse:
    """An anonymized response."""

    filename: str
    content: Optional[Union[List, Dict, str]] = None


@dataclass
class ChargingSettings:
    """Charging settings to control the vehicle."""

    chargingTarget: Optional[int]
    isUnlockCableActive = None
    acLimitValue: Optional[int] = None
    dcLoudness = None


class MyBMWAPIError(Exception):
    """General BMW API error."""


class MyBMWAuthError(MyBMWAPIError):
    """Auth-related error from BMW API (HTTP status codes 401 and 403)."""


class MyBMWQuotaError(MyBMWAPIError):
    """Quota exceeded on BMW API."""


class MyBMWRemoteServiceError(MyBMWAPIError):
    """Error when executing remote services."""



"""Models the state of a vehicle."""

import datetime
import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from bimmer_connected.const import ATTR_ATTRIBUTES, ATTR_STATE
from bimmer_connected.models import StrEnum, ValueWithUnit, VehicleDataBase
from bimmer_connected.utils import parse_datetime

_LOGGER = logging.getLogger(__name__)


class ConditionBasedServiceStatus(StrEnum):
    """Status of the condition based services."""

    OK = "OK"
    OVERDUE = "OVERDUE"
    PENDING = "PENDING"
    UNKNOWN = "UNKNOWN"


@dataclass
class ConditionBasedService:
    """Entry in the list of condition based services."""

    service_type: str
    state: ConditionBasedServiceStatus
    due_date: Optional[datetime.datetime]
    due_distance: ValueWithUnit

    @classmethod
    def from_api_entry(
        cls,
        type: str,
        status: str,
        dateTime: Optional[str] = None,
        mileage: Optional[int] = None,
        **kwargs,
    ):
        """Parse a condition based service entry from the API format to `ConditionBasedService`."""
        due_distance = ValueWithUnit(mileage, "km") if mileage else ValueWithUnit(None, None)
        due_date = parse_datetime(dateTime) if dateTime else None
        return cls(type, ConditionBasedServiceStatus(status), due_date, due_distance)


@dataclass
class ConditionBasedServiceReport(VehicleDataBase):
    """Parse and summarizes condition based services (e.g. next oil service)."""

    messages: List[ConditionBasedService] = field(default_factory=list)
    """List of the condition based services."""

    is_service_required: bool = False
    """Indicate if a service is required."""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parse doors and windows."""
        retval: Dict[str, Any] = {}

        if ATTR_STATE in vehicle_data and "requiredServices" in vehicle_data[ATTR_STATE]:
            messages = vehicle_data[ATTR_STATE]["requiredServices"]
            retval["messages"] = [ConditionBasedService.from_api_entry(**m) for m in messages]
            retval["is_service_required"] = any((m.state != ConditionBasedServiceStatus.OK) for m in retval["messages"])

        return retval


class CheckControlStatus(StrEnum):
    """Status of the condition based services."""

    OK = "OK"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    CRITICAL = "CRITICAL"
    UNKNOWN = "UNKNOWN"


@dataclass
class CheckControlMessage:
    """Check control message sent from the server."""

    description_short: str
    description_long: Optional[str]
    state: CheckControlStatus

    @classmethod
    def from_api_entry(cls, type: str, severity: str, longDescription: Optional[str] = None, **kwargs):
        """Parse a check control entry from the API format to `CheckControlMessage`."""
        return cls(type, longDescription, CheckControlStatus(severity))


@dataclass
class CheckControlMessageReport(VehicleDataBase):
    """Parse and summarizes check control messages (e.g. low tire pressure)."""

    messages: List[CheckControlMessage] = field(default_factory=list)
    """List of check control messages."""

    has_check_control_messages: bool = False
    """Indicate if check control messages are present."""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parse doors and windows."""
        retval: Dict[str, Any] = {}

        if ATTR_STATE in vehicle_data and "checkControlMessages" in vehicle_data[ATTR_STATE]:
            messages = vehicle_data[ATTR_STATE]["checkControlMessages"]
            retval["messages"] = [CheckControlMessage.from_api_entry(**m) for m in messages if m["severity"] != "OK"]
            retval["has_check_control_messages"] = len([m for m in retval["messages"] if m.state != "LOW"]) > 0

        return retval


@dataclass
class Headunit(VehicleDataBase):
    """Parse and summarizes headunit hard/software versions."""

    idrive_version: str = ""
    """IDRIVE generation."""

    headunit_type: str = ""
    """Type of headunit."""

    software_version: str = ""
    """Current software revision of vehicle"""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parse headunit hard/software."""
        retval: Dict[str, Any] = {}

        if ATTR_ATTRIBUTES in vehicle_data and "softwareVersionCurrent" in vehicle_data[ATTR_ATTRIBUTES]:
            retval["idrive_version"] = vehicle_data[ATTR_ATTRIBUTES]["hmiVersion"]
            retval["headunit_type"] = vehicle_data[ATTR_ATTRIBUTES]["headUnitType"]

            istep = vehicle_data[ATTR_ATTRIBUTES]["softwareVersionCurrent"]["iStep"]
            month = vehicle_data[ATTR_ATTRIBUTES]["softwareVersionCurrent"]["puStep"]["month"]
            year = vehicle_data[ATTR_ATTRIBUTES]["softwareVersionCurrent"]["puStep"]["year"]
            model_year = vehicle_data[ATTR_ATTRIBUTES]["year"]

            retval["software_version"] = f"{month:02d}/{str(model_year)[:2]}{year:02d}.{str(istep)[1:]}"

        return retval
